import serial
import time

import numpy

#
#---------------------------
# Maestro Servo Controller
#---------------------------
#
# Support for the Pololu Maestro line of servo controllers
#
# Steven Jacobs -- Aug 2013
# https://github.com/FRC4564/Maestro/
#
# These functions provide access to many of the Maestro's capabilities using the
# Pololu serial protocol
#


class Controller:
    # When connected via USB, the Maestro creates two virtual serial ports
    # /dev/ttyACM0 for commands and /dev/ttyACM1 for communications.
    # Be sure the Maestro is configured for "USB Dual Port" serial mode.
    # "USB Chained Mode" may work as well, but hasn't been tested.
    #
    # Pololu protocol allows for multiple Maestros to be connected to a single
    # serial port. Each connected device is then indexed by number.
    # This device number defaults to 0x0C (or 12 in decimal), which this module
    # assumes.  If two or more controllers are connected to different serial
    # ports, or you are using a Windows OS, you can provide the tty port.  For
    # example, '/dev/ttyACM2' or for Windows, something like 'COM3'.
    def __init__(self, port='/dev/ttyACM0', device=0x0c):
        # Open the command port
        self.usb = serial.Serial(port, timeout=0.1)
        # Command lead-in and device number are sent for each Pololu serial command.
        self.PololuCmd = chr(0xaa) + chr(device)
        # Track target position for each servo. The function isMoving() will
        # use the Target vs Current servo position to determine if movement is
        # occuring.  Upto 24 servos on a Maestro, (0-23). Targets start at 0.
        self.Targets = [0] * 24
        # Servo minimum and maximum targets can be restricted to protect components.
        self.Mins = [0] * 24
        self.Maxs = [0] * 24

    # Cleanup by closing USB serial port
    def close(self):
        self.usb.close()

    def read(self, *args, **kwargs):
        data = self.usb.read(*args, **kwargs)
        if len(data) == 0:
            raise serial.SerialTimeoutException()
        return data

    # Send a Pololu command out the serial port
    def sendCmd(self, cmd):
        cmdStr = self.PololuCmd + cmd
        self.usb.write(bytes(cmdStr, 'latin-1'))

    # Set channels min and max value range.  Use this as a safety to protect
    # from accidentally moving outside known safe parameters. A setting of 0
    # allows unrestricted movement.
    #
    # ***Note that the Maestro itself is configured to limit the range of servo travel
    # which has precedence over these values.  Use the Maestro Control Center to configure
    # ranges that are saved to the controller.  Use setRange for software controllable ranges.
    def setRange(self, chan, min, max):
        self.Mins[chan] = min
        self.Maxs[chan] = max

    # Return Minimum channel range value
    def getMin(self, chan):
        return self.Mins[chan]

    # Return Maximum channel range value
    def getMax(self, chan):
        return self.Maxs[chan]

    # Set channel to a specified target value.  Servo will begin moving based
    # on Speed and Acceleration parameters previously set.
    # Target values will be constrained within Min and Max range, if set.
    # For servos, target represents the pulse width in of quarter-microseconds
    # Servo center is at 1500 microseconds, or 6000 quarter-microseconds
    # Typcially valid servo range is 3000 to 9000 quarter-microseconds
    # If channel is configured for digital output, values < 6000 = Low ouput
    def setTarget(self, chan, target):
        # if Min is defined and Target is below, force to Min
        if self.Mins[chan] > 0 and target < self.Mins[chan]:
            target = self.Mins[chan]
        # if Max is defined and Target is above, force to Max
        if self.Maxs[chan] > 0 and target > self.Maxs[chan]:
            target = self.Maxs[chan]
        #
        target = int(round(target))
        lsb = target & 0x7f  # 7 bits for least significant byte
        msb = (target >> 7) & 0x7f  # shift 7 and take next 7 bits for msb
        cmd = chr(0x04) + chr(chan) + chr(lsb) + chr(msb)
        self.sendCmd(cmd)
        # Record Target value
        self.Targets[chan] = target

    # Set speed of channel
    # Speed is measured as 0.25microseconds/10milliseconds
    # For the standard 1ms pulse width change to move a servo between extremes, a speed
    # of 1 will take 1 minute, and a speed of 60 would take 1 second.
    # Speed of 0 is unrestricted.
    def setSpeed(self, chan, speed):
        speed = int(round(speed))
        lsb = speed & 0x7f  # 7 bits for least significant byte
        msb = (speed >> 7) & 0x7f  # shift 7 and take next 7 bits for msb
        cmd = chr(0x07) + chr(chan) + chr(lsb) + chr(msb)
        self.sendCmd(cmd)

    # Set acceleration of channel
    # This provide soft starts and finishes when servo moves to target position.
    # Valid values are from 0 to 255. 0=unrestricted, 1 is slowest start.
    # A value of 1 will take the servo about 3s to move between 1ms to 2ms range.
    def setAccel(self, chan, accel):
        accel = int(round(accel))
        lsb = accel & 0x7f  # 7 bits for least significant byte
        msb = (accel >> 7) & 0x7f  # shift 7 and take next 7 bits for msb
        cmd = chr(0x09) + chr(chan) + chr(lsb) + chr(msb)
        self.sendCmd(cmd)

    # Get the current position of the device on the specified channel
    # The result is returned in a measure of quarter-microseconds, which mirrors
    # the Target parameter of setTarget.
    # This is not reading the true servo position, but the last target position sent
    # to the servo. If the Speed is set to below the top speed of the servo, then
    # the position result will align well with the acutal servo position, assuming
    # it is not stalled or slowed.
    def getPosition(self, chan):
        cmd = chr(0x10) + chr(chan)
        self.sendCmd(cmd)
        lsb = ord(self.read())
        msb = ord(self.read())
        return (msb << 8) + lsb

    # Test to see if a servo has reached the set target position.  This only provides
    # useful results if the Speed parameter is set slower than the maximum speed of
    # the servo.  Servo range must be defined first using setRange. See setRange comment.
    #
    # ***Note if target position goes outside of Maestro's allowable range for the
    # channel, then the target can never be reached, so it will appear to always be
    # moving to the target.
    def isMoving(self, chan):
        if self.Targets[chan] > 0:
            if self.getPosition(chan) != self.Targets[chan]:
                return True
        return False

    def wait(self, chan):
        while self.isMoving(chan):
            time.sleep(1 / 10)

    # Have all servo outputs reached their targets? This is useful only if Speed and/or
    # Acceleration have been set on one or more of the channels. Returns True or False.
    # Not available with Micro Maestro.
    def getMovingState(self):
        cmd = chr(0x13)
        self.sendCmd(cmd)
        if self.usb.read() == chr(0):
            return False
        else:
            return True

    # Run a Maestro Script subroutine in the currently active script. Scripts can
    # have multiple subroutines, which get numbered sequentially from 0 on up. Code your
    # Maestro subroutine to either infinitely loop, or just end (return is not valid).
    def runScriptSub(self, subNumber):
        cmd = chr(0x27) + chr(subNumber)
        # can pass a param with command 0x28
        # cmd = chr(0x28) + chr(subNumber) + chr(lsb) + chr(msb)
        self.sendCmd(cmd)

    # Stop the current Maestro Script
    def stopScript(self):
        cmd = chr(0x24)
        self.sendCmd(cmd)


class AngleController(Controller):
    def __init__(self, minPosition=992, maxPosition=2000,
                 minAngle=numpy.radians(-45), maxAngle=numpy.radians(45), **kwargs):
        super().__init__(**kwargs)

        self.angleScale = \
            (maxAngle - minAngle) / (maxPosition - minPosition)
        self.angleOffset = minAngle - minPosition * self.angleScale

    def fromAngle(self, angle):
        position = (angle - self.angleOffset) / self.angleScale
        return position

    def toAngle(self, position):
        angle = position * self.angleScale + self.angleOffset
        return angle

    def setAngleRange(self, chan, minAngle, maxAngle):
        minPosition = self.fromAngle(minAngle)
        maxPosition = self.fromAngle(maxAngle)
        self.setRange(chan, 4 * minPosition, 4 * maxPosition)

    def getAngleMin(self, chan):
        minPosition = self.getMin(chan) / 4
        return self.toAngle(self.minPosition)

    def getAngleMax(self, chan):
        maxPosition = self.getMax(chan) / 4
        return self.toAngle(maxPosition)

    def setAngleTarget(self, chan, angleTarget):
        position = self.fromAngle(angleTarget)
        self.setTarget(chan, 4 * position)

    def getAngle(self, chan):
        position = self.getPosition(chan) / 4
        return self.toAngle(position)



import collections
import atexit
import enum
import warnings

import cepton_sdk.c
from cepton_sdk.common import *

__all__ = [
    "ControlFlag",
    "FrameMode",
    "deinitialize",
    "get_control_flags",
    "get_frame_length",
    "get_frame_mode",
    "get_port",
    "has_control_flag",
    "is_initialized",
    "SensorError",
    "set_control_flags",
    "set_frame_options",
    "set_port",
]

SensorError = collections.namedtuple(
    "SensorError", "error_code message error_data")


@enum.unique
class ControlFlag(enum.IntEnum):
    DISABLE_NETWORK = 1 << 1
    DISABLE_IMAGE_CLIP = 1 << 2
    DISABLE_DISTANCE_CLIP = 1 << 3
    ENABLE_MULTIPLE_RETURNS = 1 << 4


@enum.unique
class FrameMode(enum.IntEnum):
    STREAMING = 0
    TIMED = 1
    COVER = 2
    CYCLE = 3


class _Manager(object):
    def __init__(self):
        _error_callback = None
        _c_on_error = None

    def initialize(self, control_flags=None, error_callback=None,
                   frame_mode=None, frame_length=None, port=None):
        if error_callback is not None:
            self._error_callback = error_callback
        self._c_on_error = \
            cepton_sdk.c.C_SensorErrorCallback(self._global_on_error)

        c_options = cepton_sdk.c.c_create_options()
        if control_flags is not None:
            c_options.control_flags = control_flags
        if frame_length is not None:
            c_options.frame.length = frame_length
        if frame_mode is not None:
            c_options.frame.mode = frame_mode
        if port is not None:
            c_options.port = port
        cepton_sdk.c.c_initialize(
            cepton_sdk.c.SDK_VERSION, c_options, self._c_on_error, None)

    def deinitialize(self):
        try:
            cepton_sdk.c.c_deinitialize()
        except BaseException as e:
            traceback.print_exc()

        self._callback = None

    def set_error_callback(self, error_callback):
        self._error_callback = error_callback

    def clear_error_callback(self):
        self._error_callback = None

    @staticmethod
    def _global_on_error(*args):
        _manager._on_error(*args)

    def _on_error(self, sensor_handle, error_code, message, error_data, error_data_size, user_data):
        cepton_sdk.c.check_error_code(error_code)
        if error_code:
            return

        if self._error_callback is not None:
            error = SensorError(
                error_code=error_code, message=message, error_data=None)
            self._error_callback(sensor_handle, error)


_manager = _Manager()


def is_initialized():
    return bool(cepton_sdk.c.c_is_initialized())


deinitialize = _manager.deinitialize

set_error_callback = _manager.set_error_callback
clear_error_callback = _manager.clear_error_callback


def get_control_flags():
    return int(cepton_sdk.c.c_get_control_flags())


def set_control_flags(mask, flags):
    cepton_sdk.c.c_set_control_flags(mask, flags)


def has_control_flag(flag):
    return bool(cepton_sdk.c.c_has_control_flag(flag))


def get_port():
    return int(cepton_sdk.c.c_get_port())


def set_port(port):
    cepton_sdk.c.c_set_port(port)


def set_frame_options(mode, length=None):
    c_options = cepton_sdk.c.c_create_frame_options()
    if length is not None:
        c_options.length = length
    cepton_sdk.c.c_set_frame_options(c_options)


def get_frame_length():
    return float(cepton_sdk.c.c_get_frame_length())


def get_frame_mode():
    return FrameMode(cepton_sdk.c.c_get_frame_mode())



__author__ = "Cepton Technologies"
__version__ = "0.1.1"

import cepton_sdk.capture_replay

from cepton_sdk.api import *
from cepton_sdk.c import C_ErrorCode, C_Error, C_Warning
from cepton_sdk.core import *
from cepton_sdk.point import *
from cepton_sdk.sensor import *



import ctypes
import os.path

import cepton_sdk.c
import cepton_sdk.listener

__all__ = [
    "get_length",
    "get_position",
    "get_start_time",
    "get_time",
    "is_end",
    "is_open",
    "rewind",
    "seek_relative",
    "seek",
]


def is_open():
    return bool(cepton_sdk.c.c_capture_replay_is_open())


def open(capture_path):
    capture_path = capture_path.encode("UTF-8")
    cepton_sdk.c.c_capture_replay_open(capture_path)
    cepton_sdk.listener.clear_cache()


def close():
    try:
        cepton_sdk.c.c_capture_replay_close()
    except IOError as e:
        warnings.warn(e)
    cepton_sdk.listener.clear_cache()


def get_start_time():
    return 1e-6 * float(cepton_sdk.c.c_capture_replay_get_start_time())


def get_position():
    return float(cepton_sdk.c.c_capture_replay_get_position())


def get_time():
    return get_start_time() + get_position()


def get_length():
    return float(cepton_sdk.c.c_capture_replay_get_length())


def is_end():
    return bool(cepton_sdk.c.c_capture_replay_is_end())


def rewind():
    cepton_sdk.c.c_capture_replay_rewind()
    cepton_sdk.listener.clear_cache()


def seek(t):
    cepton_sdk.c.c_capture_replay_seek(t)
    cepton_sdk.listener.clear_cache()


def seek_relative(t):
    t += get_position()
    seek(t)


def get_enable_loop():
    return bool(cepton_sdk.c.c_capture_replay_get_enable_loop())


def set_enable_loop(enable_loop):
    cepton_sdk.c.c_capture_replay_set_enable_loop(enable_loop)


def resume_blocking(t_length=None):
    if t_length is None:
        cepton_sdk.c.c_capture_replay_resume_blocking_once()
    else:
        cepton_sdk.c.c_capture_replay_resume_blocking(t_length)


def is_running():
    return bool(cepton_sdk.c.c_capture_replay_is_running())


def resume():
    cepton_sdk.c.c_capture_replay_resume()


def pause():
    cepton_sdk.c.c_capture_replay_pause()



import datetime
import time

import cepton_sdk.c
import cepton_sdk.capture_replay
import cepton_sdk.core
import cepton_sdk.listener
import cepton_sdk.sensor
from cepton_sdk.common import *

__all__ = [
    "clear_cache",
    "get_image_frames",
    "get_image_points",
    "get_sensor_image_frames",
    "get_sensor_image_points_by_n",
    "get_sensor_image_points",
    "get_sensors",
    "get_time",
    "initialize",
    "get_timestamp",
    "is_end",
    "is_live",
    "Sensor",
    "wait",
]


def is_live():
    """Returns true if capture replay is not open, false otherwise."""
    return not cepton_sdk.capture_replay.is_open()


def is_end():
    """Returns true if capture replay is open and at end, false otherwise."""
    if is_live():
        return False  # Live sensors never end
    if cepton_sdk.capture_replay.get_enable_loop():
        return False  # Looping captures never end
    return cepton_sdk.capture_replay.is_end()


def get_timestamp():
    """Returns unix timestamp"""
    return datetime.datetime.utcnow().timestamp()


def get_time():
    """Returns capture replay time or live time"""
    if is_live():
        return get_timestamp()
    else:
        return cepton_sdk.capture_replay.get_time()


def wait(t_length=0.1):
    """Resumes capture replay or sleeps for duration."""
    if cepton_sdk.capture_replay.is_open():
        cepton_sdk.capture_replay.resume_blocking(t_length)
    else:
        time.sleep(t_length)


def _wait_on_func(func, timeout=None):
    if timeout is not None:
        t_start = time.time()
    while not func():
        wait()
        if timeout is not None:
            if (time.time() - t_start) > timeout:
                raise RuntimeError("timed out")


def initialize(capture_path=None, control_flags=0, error_callback=None, port=None, **kwargs):
    """Initializes SDK. Optionally starts capture replay.

    Arguments:
        control_flags: :class:`cepton_sdk.ControlFlag`
    """
    if capture_path is not None:
        control_flags |= cepton_sdk.core.ControlFlag.DISABLE_NETWORK
    options = {
        "control_flags": control_flags,
        "frame_mode": cepton_sdk.core.FrameMode.COVER,
    }
    options.update(kwargs)
    if error_callback is not None:
        options["error_callback"] = error_callback
    if port is not None:
        options["port"] = port
    cepton_sdk.core._manager.initialize(**options)
    cepton_sdk.listener.initialize()

    if capture_path is None:
        time.sleep(1)
    else:
        cepton_sdk.capture_replay.open(capture_path)
        cepton_sdk.capture_replay.resume_blocking(1)
        cepton_sdk.capture_replay.seek(0)


def clear_cache():
    cepton_sdk.c.c_clear_cache()
    cepton_sdk.listener.clear_cache()


def _get_points_wait(wait_func, return_partial=False, timeout=None):
    if not cepton_sdk.core.is_initialized():
        raise cepton_sdk.c.C_Error(cepton_sdk.c.C_Error.CEPTON_ERROR_NOT_INITIALIZED)

    if wait_func():
        return
    if is_end():
        error_code = cepton_sdk.c.C_ErrorCode.CEPTON_ERROR_EOF
        raise cepton_sdk.c.C_Error.create(error_code)
    try:
        _wait_on_func(wait_func, timeout=timeout)
    except cepton_sdk.c.C_Error as e:
        if return_partial:
            if e.error_code == cepton_sdk.c.C_ErrorCode.CEPTON_ERROR_EOF:
                return
        raise


def _get_frames(listener, **kwargs):
    def wait_func(): return listener.has_points()
    _get_points_wait(wait_func, **kwargs)
    return listener.get_points()


@static_vars(listener=cepton_sdk.listener.PointsListener())
def get_image_frames(**kwargs):
    """Returns next frames of points for all sensors.

    Returns:
        Dictionary of lists of image points, indexed by serial number.
    """
    return _get_frames(get_image_frames.listener, **kwargs)


@static_vars(listeners={})
def get_sensor_image_frames(serial_number, **kwargs):
    """Returns next frames of points for specified sensor.

    Args:
        serial_number: Sensor serial number.
    Returns:
        List of image points.
    """
    listener = _get_sensor_listener(
        get_sensor_image_frames.listeners, serial_number,
        cepton_sdk.listener.SensorPointsListener)
    return _get_frames(listener, **kwargs)


def _get_points(listener, frame_length, **kwargs):
    if frame_length is None:
        if is_live():
            raise ValueError("must specify frame lenghth for live sensor")
        kwargs["return_partial"] = True

        def wait_func():
            listener.update()
            return False
    else:
        def wait_func(): return listener.has_points(frame_length)
    _get_points_wait(wait_func, **kwargs)
    if frame_length is None:
        listener.update()
        return listener.get_points_by_t(listener.t_max)
    else:
        return listener.get_points(frame_length)


@static_vars(listener=cepton_sdk.listener.TimeAccumulatedPointsListener(t_latency=0.1))
def get_image_points(frame_length, return_partial=False, timeout=None):
    """Returns next chunk of points for all sensors.

    If using capture replay and frame length is None, returns all remaining
    points in capture file.

    Args:
        frame_length: length of time chunk [sec].
    Returns:
        Dictionary of image points, indexed by serial number.
    """
    return _get_points(
        get_image_points.listener, frame_length, return_partial=return_partial,
        timeout=timeout)


def _get_sensor_listener(listeners, serial_number, cls, *args, **kwargs):
    if serial_number not in listeners:
        listeners[serial_number] = cls(serial_number, *args, **kwargs)
    return listeners[serial_number]


@static_vars(listeners={})
def get_sensor_image_points(serial_number, frame_length, return_partial=False, timeout=None):
    """Returns next chunk of points for specified sensor.

    If using capture replay and frame length is None, returns all remaining
    points in capture file.

    Args:
        serial_number: Sensor serial number.
        frame_length: Length of time chunk [sec].
    Returns:
        Image points.
    """
    listener = _get_sensor_listener(
        get_sensor_image_points.listeners, serial_number,
        cepton_sdk.listener.TimeAccumulatedSensorPointsListener)
    return _get_points(
        listener, frame_length, return_partial=return_partial, timeout=timeout)


def _get_points_by_n(listener, n_points, **kwargs):
    def wait_func(): return listener.has_points(n_points)
    _get_points_wait(wait_func, **kwargs)
    return listener.get_points(n_points)


@static_vars(listeners={})
def get_sensor_image_points_by_n(serial_number, n_points, **kwargs):
    """Returns next chunk of points for specified sensor.

    Args:
        serial_number: Sensor serial number.
        n_points: Number of points.
    Returns:
        Image points.
    """
    listener = _get_sensor_listener(
        get_sensor_image_points_by_n.listeners, serial_number,
        cepton_sdk.listener.NAccumulatedSensorPointsListener)
    return _get_points_by_n(listener, n_points, **kwargs)


class Sensor(object):
    """
    Attributes:
        information (:class:`cepton_sdk.SensorInformation`)
    """

    def __init__(self, sensor_info):
        if type(sensor_info) is dict:
            self.information = SensorInformation.from_dict(sensor_info)
        else:
            self.information = sensor_info

    @property
    def handle(self):
        return self.information.handle

    @property
    def serial_number(self):
        return self.information.serial_number

    @property
    def model_name(self):
        return self.information.model_name

    @property
    def firmware_version(self):
        return self.information.firmware_version

    @property
    def model(self):
        return self.information.model  # This is SensorModel enum

    @classmethod
    def create_by_index(cls, sensor_index):
        sensor_info = \
            cepton_sdk.sensor.get_sensor_information_by_index(sensor_index)
        return cls(sensor_info)

    @classmethod
    def create_by_handle(cls, sensor_handle):
        sensor_info = \
            cepton_sdk.sensor.get_sensor_information_by_handle(sensor_handle)
        return cls(sensor_info)

    @classmethod
    def create(cls, sensor_serial_number):
        sensor_info = \
            cepton_sdk.sensor.get_sensor_information(sensor_serial_number)
        return cls(sensor_info)

    def update(self):
        """Update sensor information.

        Should be called often, to pull latest sensor information.
        """
        self.information = \
            cepton_sdk.sensor.get_sensor_information_by_handle(self.handle)

    def get_image_frames(self, *args, **kwargs):
        """See `cepton_sdk.get_sensor_image_frames`"""
        return get_sensor_image_frames(self.serial_number, *args, **kwargs)

    def get_image_points(self, *args, **kwargs):
        """See `cepton_sdk.get_sensor_image_points`"""
        return get_sensor_image_points(self.serial_number, *args, **kwargs)

    def get_image_points_by_n(self, *args, **kwargs):
        """See `cepton_sdk.get_sensor_image_points_by_n`"""
        return get_sensor_image_points_by_n(self.serial_number, *args, **kwargs)

    def is_hr80t(self):
        return self.model == cepton_sdk.sensor.SensorModel.HR80T

    def is_hr80t_r2(self):
        return self.model == cepton_sdk.sensor.SensorModel.HR80T_R2

    def is_hr80w(self):
        return self.model == cepton_sdk.sensor.SensorModel.HR80W

    def is_sora(self):
        return self.model == cepton_sdk.sensor.SensorModel.SORA_200

    def is_vista860(self):
        return self.model == cepton_sdk.sensor.SensorModel.VISTA_860


def get_sensors(cls=Sensor):
    """Returns attached sensors.

    Returns:
        Dictionary of sensors, indexed by serial number.
    """
    sensors_dict = {}
    for i_sensor in range(cepton_sdk.sensor.get_n_sensors()):
        sensor = cls.create_by_index(i_sensor)
        sensors_dict[sensor.serial_number] = sensor
    return sensors_dict



import argparse
import datetime
import os
import os.path
import signal
import subprocess
import sys
import threading
import time

import PyQt5.QtWidgets
import serial
import serial.tools.list_ports

__all__ = [
    "add_execute_command_arguments",
    "add_output_path_arguments",
    "ArgumentParserMixin",
    "create_directory",
    "execute_command",
    "fix_path",
    "get_environment",
    "get_output_path",
    "get_serial_port_dialog",
    "get_timestamp_str",
    "get_timestamp",
    "modify_path",
    "parse_execute_command_arguments",
    "parse_time_hms",
    "remove_extension",
    "run_background",
    "set_extension",
    "wait_on_background",
]


def get_timestamp():
    return datetime.datetime.utcnow().timestamp()


def get_timestamp_str():
    return datetime.datetime.now().strftime("%y%m%d_%H%M%S")


def parse_time_hms(s):
    """Convert HH:MM:SS to seconds"""
    parts = [float(x) for x in s.split(":")]
    sec = 0
    for i, part in enumerate(reversed(parts)):
        sec += part * 60**i
    return sec


STRING_BOOL_LOOKUP = {
    "true": True,
    "false": False,
}


def get_environment(key):
    s = os.environ[key]
    if s.lower() in STRING_BOOL_LOOKUP:
        return STRING_BOOL_LOOKUP[s.lower()]
    return s


def get_serial_port_dialog(filter_func=None):
    app = PyQt5.QtWidgets.QApplication([])

    ports = serial.tools.list_ports.comports()
    if filter_func is not None:
        ports = list(filter(filter_func, ports))
    port_names = [str(x) for x in ports]
    port_lookup = {str(x): x.device for x in ports}
    item, valid = PyQt5.QtWidgets.QInputDialog.getItem(
        None, "Select Serial Port", "Port:", port_names, 0, False)
    if valid:
        return port_lookup[item]
    else:
        return None

    app.exec_()

# ------------------------------------------------------------------------------
# Execute process
# ------------------------------------------------------------------------------


__local = {}
__local["procs"] = []
__local["threads"] = []


class BackgroundProcess(object):
    def __init__(self, *args, **kwargs):
        self.proc = subprocess.Popen(*args, **kwargs)

    def __del__(self):
        self.proc.terminate()
        for i in range(10):
            if self.proc.poll() is not None:
                return
            time.sleep(0.1)
        self.proc.kill()


class BackgroundThread(object):
    def __init__(self, *args, **kwargs):
        self.shutdown_event = threading.Event()
        kwargs["kwargs"]["shutdown_event"] = self.shutdown_event
        self.thread = threading.Thread(*args, **kwargs)

    def __del__(self):
        self.shutdown_event.set()
        self.thread.join(1.0)


def wait_on_background():
    for proc in __local["procs"]:
        proc.proc.wait()
    __local["procs"] = []
    for thread in __local["threads"]:
        thread.thread.join()
    __local["threads"] = []


def run_background(func, args=(), kwargs={}):
    thread = BackgroundThread(target=func, args=args, kwargs=kwargs)
    thread.thread.start()
    __local["threads"].append(thread)


def execute_command(cmd_list, background=False, quiet=False, **kwargs):
    options = {
        # "check": True,
    }
    if quiet:
        options.update({
            "stdout": subprocess.DEVNULL,
            "stderr": subprocess.DEVNULL,
        })
    options.update(kwargs)
    if background:
        proc = BackgroundProcess(cmd_list, **options)
        __local["procs"].append(proc)
        return proc.proc
    else:
        subprocess.run(cmd_list, **options)


def add_execute_command_arguments(parser):
    parser.add_argument("-q", "--quiet", action="store_true")


def parse_execute_command_arguments(args):
    return {
        "quiet": args.quiet,
    }


# ------------------------------------------------------------------------------
# Path
# ------------------------------------------------------------------------------
def create_directory(path):
    if not os.path.exists(path):
        os.makedirs(path)


def fix_path(path):
    path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    path = os.path.realpath(path)
    path = os.path.abspath(path)
    return path


def remove_extension(path):
    return os.path.splitext(path)[0]


def set_extension(path, new_ext):
    return os.path.splitext(path)[0] + new_ext


def modify_path(path, new_ext=None, prefix="", postfix=""):
    head, basename = os.path.split(path)
    name, ext = os.path.splitext(path)
    if new_ext is not None:
        ext = new_ext
    basename = "{}{}{}{}".format(prefix, name, postfix, ext)
    path = os.path.join(head, basename)
    return path


def add_output_path_arguments(parser, prefix="", postfix="_processed"):
    parser.add_argument("-o", "--output")
    parser.add_argument("--prefix", default=prefix)
    parser.add_argument("--postfix", default=postfix)


def get_output_path(input_path, args):
    if args.output is None:
        options = {
            "prefix": args.prefix,
            "postfix": args.postfix,
        }
        return modify_path(input_path, **options)
    else:
        return fix_path(args.output)

# ------------------------------------------------------------------------------
# Arguments
# ------------------------------------------------------------------------------


class ArgumentParserMixin(object):
    @classmethod
    def add_arguments(cls, parser):
        raise NotImplementedError()

    @classmethod
    def parse_arguments(cls, args):
        raise NotImplementedError()

    @classmethod
    def from_arguments(cls, parser_args, *args, **kwargs):
        options = cls.parse_arguments(parser_args)
        options.update(kwargs)
        return cls(*args, **options)



"""SORA scan demo with Polulu Micro Maestro."""

import re
import sys
import time

import numpy
import serial
import serial.tools.list_ports
import transforms3d.quaternions

import cepton_util.demo.maestro

# ------------------------------------------------------------------------------
# Config
# ------------------------------------------------------------------------------

"""Minimum scan angle [degrees]. Must be in range [-45, 45]."""
ANGLE_MIN = -30

"""Maximum scan angle [degrees]. Must be in range [-45, 45]."""
ANGLE_MAX = 30

"""Scan speed. Must be in range [1, 10]."""
SPEED = 3

# ------------------------------------------------------------------------------

CHANNEL = 0


def main():
    # Get port by vid, pid
    def filter_func(x): return ((x.vid == 8187) and (x.pid == 137))
    ports = serial.tools.list_ports.comports()
    ports = list(filter(filter_func, ports))
    connected = False
    for port in ports:
        controller = cepton_util.demo.maestro.AngleController(port=port.device)
        try:
            controller.getPosition(CHANNEL)
        except serial.SerialTimeoutException:
            continue
        connected = True
        break

    if not connected:
        raise IOError("Maestro not connected!")

    controller.setSpeed(CHANNEL, SPEED)
    while True:
        for target in [ANGLE_MIN, ANGLE_MAX]:
            target = numpy.radians(target)
            controller.setAngleTarget(CHANNEL, target)
            while controller.isMoving(CHANNEL):
                angle = controller.getAngle(CHANNEL)

                q_1 = transforms3d.quaternions.axangle2quat(
                    [0, 1, 0], -numpy.pi / 2)
                q_2 = transforms3d.quaternions.axangle2quat([0, 0, 1], -angle)
                q = transforms3d.quaternions.qmult(q_2, q_1)

                print("{} {} {} {} 0 0 0".format(q[0], q[1], q[2], q[3]))
                time.sleep(0.1)


if __name__ == "__main__":
    main()



import collections
import threading

import cepton_sdk.c
import cepton_sdk.point
import cepton_sdk.sensor
from cepton_sdk.common import *


class _Frame(object):
    def __init__(self, frame_id, sensor_serial_number, points):
        self.id = frame_id
        self.sensor_serial_number = sensor_serial_number
        self.points = points


class _FramesListenerBase(object):
    """Global points listener class.

    Stores history of most recently received frames.
    Use PointsListener and SensorPointsListener for higher level api.
    """

    _instances = []

    def __init__(self):
        self._instances.append(self)
        self._lock = threading.Lock()
        self._reset_id = 0
        self._frame_id = 0
        self._frames = collections.deque()
        self._n_points = 0
        self._max_n_points = int(1e6)

    def initialize(self):
        raise NotImplementedError()

    def reset(self):
        with self._lock:
            self._reset_id += 1
            self._frames.clear()
            self._n_points = 0

    @property
    def reset_id(self):
        with self._lock:
            return self._reset_id

    @property
    def frame_id(self):
        with self._lock:
            return self._frame_id

    def _is_valid_frame(self, frame, sensor_serial_number=None, frame_id_lb=None):
        if sensor_serial_number is not None:
            if (frame.sensor_serial_number != sensor_serial_number):
                return False
        if frame_id_lb is not None:
            if frame.id < frame_id_lb:
                return False
        return True

    def has_frames(self, **kwargs):
        with self._lock:
            for frame in reversed(self._frames):
                if self._is_valid_frame(frame, **kwargs):
                    return True
        return False

    def get_frames(self, **kwargs):
        with self._lock:
            frames_list = \
                [frame for frame in self._frames
                 if self._is_valid_frame(frame, **kwargs)]
        return frames_list

    def _add_points(self, sensor_handle, points):
        sensor_info = \
            cepton_sdk.sensor.get_sensor_information_by_handle(sensor_handle)
        with self._lock:
            frame = \
                cepton_sdk.listener._Frame(
                    self._frame_id, sensor_info.serial_number, points)
            self._frame_id += 1
            self._frames.append(frame)
            self._n_points += len(frame.points)
            while self._n_points > self._max_n_points:
                frame_tmp = self._frames.popleft()
                self._n_points -= len(frame_tmp.points)

def _global_on_image_points(*args):
    _frames_listener._on_image_points(*args)


class _FramesListener(_FramesListenerBase):
    def initialize(self):
        self._c_on_points = \
            cepton_sdk.c.C_SensorImageDataCallback(_global_on_image_points)
        cepton_sdk.c.c_listen_image_frames(self._c_on_points, None)

    def _on_image_points(self, sensor_handle, n_points, c_image_points_ptr, c_user_data_ptr):
        image_points = \
            cepton_sdk.point.ImagePoints.from_c(n_points, c_image_points_ptr)
        self._add_points(sensor_handle, image_points)


_frames_listener = _FramesListener()


def initialize():
    for frames_listener in _FramesListenerBase._instances:
        frames_listener.initialize()


def clear_cache():
    for frames_listener in _FramesListenerBase._instances:
        frames_listener.reset()


class _ListenerBase(object):
    def __init__(self, frames_listener=_frames_listener):
        self._frames_listener = frames_listener
        self.reset()

    def reset(self):
        self._frame_id = self._frames_listener.frame_id
        self._reset_id = self._frames_listener.reset_id

    def update(self):
        if self._reset_id != self._frames_listener.reset_id:
            self.reset()


class PointsListener(_ListenerBase):
    """Listener for getting all sensor frames."""

    def has_points(self):
        self.update()
        options = {
            "frame_id_lb": self._frame_id,
        }
        return self._frames_listener.has_frames(**options)

    def get_points(self):
        self.update()
        frames_list = self._frames_listener.get_frames(frame_id_lb=self._frame_id)
        if len(frames_list) == 0:
            return {}

        self._frame_id = frames_list[-1].id + 1

        points_dict = collections.defaultdict(list)
        for frame in frames_list:
            points_dict[frame.sensor_serial_number].append(frame.points)
        return points_dict


class SensorPointsListener(_ListenerBase):
    def __init__(self, sensor_serial_number, **kwargs):
        self.sensor_serial_number = sensor_serial_number
        super().__init__(**kwargs)

    def has_points(self):
        self.update()
        return self._frames_listener.has_frames(
            frame_id_lb=self._frame_id,
            sensor_serial_number=self.sensor_serial_number)

    def get_points(self):
        self.update()
        frames = self._frames_listener.get_frames(
            frame_id_lb=self._frame_id,
            sensor_serial_number=self.sensor_serial_number)
        if not frames:
            return []
        self._frame_id = frames[-1].id + 1
        points_list = [frame.points for frame in frames]
        return points_list


class _NAccumulatedPointsListenerMixin(object):
    def has_points(self, n_points):
        self.update()
        return self._accumulator.has_points(n_points)

    def get_points(self, n_points):
        self.update()
        return self._accumulator.get_points(n_points)


class _TimeAccumulatedPointsListenerMixin(object):
    @property
    def t(self):
        return self._accumulator.t

    @property
    def t_max(self):
        return self._accumulator.t_max

    def has_points_by_t(self, t):
        self.update()
        return self._accumulator.has_points_by_t(t)

    def has_points(self, t_length):
        self.update()
        return self._accumulator.has_points(t_length)

    def get_points_by_t(self, t):
        self.update()
        return self._accumulator.get_points_by_t(t)

    def get_points(self, t_length):
        self.update()
        return self._accumulator.get_points(t_length)


class _MultipleAccumulatedPointsListenerBase(_ListenerBase):
    def __init__(self, frames_listener=_frames_listener):
        self._listener = PointsListener(frames_listener=frames_listener)
        super().__init__(frames_listener=frames_listener)

    def update(self):
        super().update()
        if not self._listener.has_points():
            return
        points_dict = self._listener.get_points()
        for sensor_serial_number, points_list in points_dict.items():
            points = points_list[0].combine(points_list)
            self._accumulator.add_points(sensor_serial_number, points)

    def reset(self):
        super().reset()
        self._accumulator.reset()
        self._listener.reset()


class _SensorAccumulatedPointsListenerBase(_ListenerBase):
    def __init__(self, sensor_serial_number, frames_listener=_frames_listener):
        self._listener = \
            cepton_sdk.listener.SensorPointsListener(
                sensor_serial_number, frames_listener=frames_listener)
        super().__init__(frames_listener=frames_listener)

    def update(self):
        super().update()

        if not self._listener.has_points():
            return
        points_list = self._listener.get_points()
        if len(points_list) == 0:
            return
        points = points_list[0].combine(points_list)
        self._accumulator.add_points(points)

    def reset(self):
        super().reset()
        self._accumulator.reset()
        self._listener.reset()


class NAccumulatedSensorPointsListener(_SensorAccumulatedPointsListenerBase,
                                       _NAccumulatedPointsListenerMixin):
    def __init__(self, sensor_serial_number, frames_listener=_frames_listener):
        self._accumulator = cepton_sdk.point.NPointsAccumulator()
        super().__init__(sensor_serial_number, frames_listener=frames_listener)


class TimeAccumulatedPointsListener(_MultipleAccumulatedPointsListenerBase,
                                    _TimeAccumulatedPointsListenerMixin):
    def __init__(self, frames_listener=_frames_listener, **kwargs):
        self._accumulator = \
            cepton_sdk.point.MultipleTimePointsAccumulator(**kwargs)
        super().__init__(frames_listener=frames_listener)


class TimeAccumulatedSensorPointsListener(_SensorAccumulatedPointsListenerBase,
                                          _TimeAccumulatedPointsListenerMixin):
    def __init__(self, sensor_serial_number, frames_listener=_frames_listener, **kwargs):
        self._accumulator = cepton_sdk.point.TimePointsAccumulator(**kwargs)
        super().__init__(sensor_serial_number, frames_listener=frames_listener)



import copy
import ctypes
import functools

import numpy

def static_vars(**kwargs):
    """Add static variables to function.
    """
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate

class single_cache(object):
    def __init__(self, func):
        self.func = func
        self.result = None
        functools.update_wrapper(self, func)

    def __call__(self, *args, **kwargs):
        if self.result is None:
            self.result = self.func(*args, **kwargs)
        return copy.deepcopy(self.result)

class ToDictMixin(object):
    @classmethod
    def _get_dict_member_names(cls):
        raise NotImplementedError()

    @classmethod
    def _from_dict_value(cls, member_name, dict_value):
        value = dict_value
        return value

    def _to_dict_value(self, member_name, value):
        dict_value = value
        return dict_value

    def update_from_dict(self, d, deep_copy=True, member_names=None):
        if member_names is None:
            member_names = list(d.keys())

        for member_name in member_names:
            dict_value = d[member_name]
            value = self._from_dict_value(member_name, dict_value)
            if deep_copy:
                value = copy.deepcopy(value)
            setattr(self, member_name, value)

    @classmethod
    def from_dict(cls, d, **kwargs):
        obj = cls()
        obj.update_from_dict(d, **kwargs)
        return obj

    def to_dict(self, deep_copy=True, member_names=None):
        if member_names is None:
            member_names = self._get_dict_member_names()

        d = {}
        for member_name in member_names:
            value = getattr(self, member_name)
            dict_value = self._to_dict_value(member_name, value)
            if deep_copy:
                dict_value = copy.deepcopy(dict_value)
            d[member_name] = dict_value
        return d

class C_Field(object):
    def __init__(self, field_name, field_type, field_width=None):
        self.name = field_name
        self.type = field_type
        self.width = field_width

    @classmethod
    def from_description(cls, descr):
        if len(descr) == 3:
            return cls(descr[0], descr[1], descr[2])
        else:
            return cls(descr[0], descr[1])

def get_c_members(c_cls):
    return {x[0]: C_Field.from_description(x) for x in c_cls._fields_ if x[0]}


def get_c_member_names(c_cls):
    return [x[0] for x in c_cls._fields_ if x[0]]


class ToCMixin(object):
    @classmethod
    def _get_c_class(cls):
        raise NotImplementedError()

    @classmethod
    def _from_c_value(cls, member_name, c_value):
        # TODO
        # c_member = self._get_c_member(member_name)
        # if issubclass(c_member.type, ctypes.Array):
        #     if issubclass(c_member.type._type_, c_char):
        #         value = c_value.decode("utf-8")
        #     else:
        #         value = list(c_value)
        value = c_value
        return value

    def _to_c_value(self, member_name, value):
        c_member = self._get_c_member(member_name)
        if issubclass(c_member.type, ctypes.Array):
            if issubclass(c_member.type._type_, c_char):
                if isinstance(value, str):
                    c_value = value.encode("utf-8")
                else:
                    c_value = value
            elif issubclass(c_member.type._type_, c_wchar):
                c_value = value
            else:
                c_value = c_member.type(*value)
        else:
            c_value = value
        return c_value

    @classmethod
    def __get_c_members(cls):
        return get_c_members(cls._get_c_class())

    @classmethod
    def _get_c_member_names(cls):
        return list(cls.__get_c_members().keys())

    @classmethod
    def _get_c_member(cls):
        return cls.__get_c_members()[member_name]

    @classmethod
    def from_c(cls, c_obj, deep_copy=True, member_names=None):
        if member_names is None:
            member_names = cls._get_c_member_names()

        obj = cls()
        for member_name in member_names:
            c_value = getattr(c_obj, member_name)
            value = cls._from_c_value(member_name, c_value)
            setattr(obj, member_name, value)
        return obj

    def to_c(self, c_type=None, deep_copy=True, member_names=None):
        if c_type is None:
            c_type = self._get_c_class()
        if member_names is None:
            member_names = self._get_c_member_names()

        c_cls = self._get_c_class()
        c_obj = c_cls()
        for member_name in member_names:
            try:
                value = getattr(self, member_name)
            except AttributeError:
                continue
            c_value = self._to_c_value(member_name, value)
            if deep_copy:
                c_value = copy.deepcopy(c_value)
            setattr(c_obj, member_name, c_value)

        return c_obj


class StructureOfArrays(object):
    """
    Group multiple arrays together and allow operations on all arrays
    simultaneously. Supports numpy operations as if it were a 1-d array.

    Avoid using instance member variables that are not listed as array members,
    since they will not be copied during operations. Instead, use class member
    variables.
    """

    def __init__(self, n=1):
        self.__n = n
        self.__checked_members = None
        self.__checked_members = \
            set(self._get_array_member_names()) | set(dir(self))

    def _check_array_members(self):
        """Check that all members are listed in `_get_array_member_names`.
        """
        all_members = set(dir(self))
        unlisted_members = all_members - self.__checked_members
        self.__checked_members |= all_members
        if len(unlisted_members) == 0:
            return
        raise RuntimeError(
            "Members not listed in `_get_array_member_names`: {}".format(
                unlisted_members))

    @classmethod
    def _get_array_member_names(cls):
        raise NotImplementedError()

    def __len__(self):
        return self.__n

    @classmethod
    def from_parent(cls, other):
        """Cast from parent class to child class.

        Copies over all parent member variables.
        """
        parent_cls = type(other)

        obj = cls(len(other))
        for name in parent_cls._get_array_member_names():
            value = getattr(other, name)
            setattr(obj, name, value)
        return obj

    def _get_indices(self, key):
        indices = numpy.arange(len(self))[key]
        indices = numpy.reshape(indices, [-1])
        return indices

    def __getitem__(self, key):
        """Supports numpy style indexing as if object were 1-d array.
        """
        indices = self._get_indices(key)

        cls = type(self)
        selected_obj = cls(indices.size)
        for name in self._get_array_member_names():
            data = getattr(self, name)
            selected_data = data[indices, ...]
            setattr(selected_obj, name, selected_data)
        return selected_obj

    def __setitem__(self, key, other):
        """Supports numpy style assignment as if object were 1-d array.
        """
        if not isinstance(other, type(self)):
            raise TypeError("Incompatible types")

        indices = self._get_indices(key)

        for name in self._get_array_member_names():
            getattr(self, name)[indices, ...] = getattr(other, name)

    @classmethod
    def combine(cls, obj_list):
        """Combine list of objects into single object.

        Concatenates all member variables along first dimension.
        """
        for obj in obj_list:
            if not isinstance(obj, cls):
                raise TypeError("Incompatible types")

        obj_list = list(obj_list)
        if len(obj_list) == 0:
            return cls(0)

        n_combined_obj = numpy.sum([len(x) for x in obj_list])
        combined_obj = cls(n_combined_obj)
        for name in obj_list[0]._get_array_member_names():
            value_list = [getattr(x, name) for x in obj_list]
            try:
                combined_value = value_list[0].combine(value_list)
            except AttributeError:
                combined_value = numpy.concatenate(value_list, axis=0)
            setattr(combined_obj, name, combined_value)
        return combined_obj

    # def _get_dtype(self, member_names=None):
    #     if member_names is None:
    #         member_names = self._get_array_member_names()
    #
    #     # TODO: deal with nested StructureOfArrays
    #     descr = []
    #     for name in member_names:
    #         value = numpy.array(getattr(self, name))
    #         assert (len(value.dtype.descr) == 1)
    #         assert (value.dtype.ndim) == 0
    #         shape_tmp = numpy.array(value.shape)[1:]
    #         if len(shape_tmp) == 0:
    #             shape_tmp = 1
    #         descr.append((name, value.dtype.name, shape_tmp))
    #     dtype = numpy.dtype(descr)
    #     return dtype
    #
    # def to_structured_array(self, member_names=None):
    #     if member_names is None:
    #         member_names = self._get_array_member_names()
    #
    #     dtype = self._get_dtype(member_names=member_names)
    #     data = numpy.zeros([len(self)], dtype=dtype)
    #     for name in member_names:
    #         data[name][...] = getattr(self, name)
    #     return data
    #
    # def update_from_structured_array(self, data, member_names=None):
    #     if member_names is None:
    #         member_names = data.dtype.names
    #
    #     dtype = self._get_dtype(member_names=member_names)
    #     for name in member_names:
    #         getattr(self, name)[...] = data[name]
    #     return self
    #
    # @classmethod
    # def from_structured_array(cls, data, member_names=None):
    #     obj = cls(data.size)
    #     obj.update_from_structured_array(data, member_names=member_names)
    #     return obj



import ctypes
import enum

import cepton_sdk.c
from cepton_sdk.common import *

__all__ = [
    "get_n_sensors",
    "get_sensor_handle",
    "get_sensor_information_by_handle",
    "get_sensor_information_by_index",
    "get_sensor_information",
    "has_sensor",
    "SensorInformation",
    "SensorModel",
]


class SensorModel(enum.Enum):
    HR80T = 1
    HR80M = 2
    HR80W = 3
    SORA_200 = 4
    VISTA_860 = 5
    HR80T_R2 = 6


class SensorInformation(ToCMixin, ToDictMixin):
    """
    Attributes:
        handle
        serial_number
        model_name
        model (:class:`cepton_sdk.SensorModel`)
        firmware_version
        last_reported_temperature
        last_reported_humidity
        last_reported_age
        gps_ts_year
        gps_ts_mont
        gps_ts_day
        gps_ts_hour
        gps_ts_min
        gps_ts_sec
        return_count
        is_mocked
        is_pps_connected
        is_nmea_connected
        is_calibrated
    """

    @classmethod
    def _get_c_class(cls):
        return cepton_sdk.c.C_SensorInformation

    @classmethod
    def _get_dict_member_names(cls):
        return cls._get_c_member_names()

    __from_c_value_functions = {
        "model_name": lambda x: x.decode("utf-8"),
        "model": SensorModel,
        "firmware_version": lambda x: x.decode("utf-8"),
        "is_mocked": bool,
        "is_pps_connected": bool,
        "is_nmea_connected": bool,
        "is_ptp_connected": bool,
        "is_calibrated": bool,
    }

    @classmethod
    def _from_c_value(cls, member_name, c_value):
        if member_name in cls.__from_c_value_functions:
            return cls.__from_c_value_functions[member_name](c_value)
        else:
            return super()._from_c_value(member_name, c_value)


def get_n_sensors():
    return int(cepton_sdk.c.c_get_n_sensors())


def get_sensor_handle(serial_number):
    c_sensor_handle = cepton_sdk.c.C_SensorHandle()
    cepton_sdk.c.c_get_sensor_handle_by_serial_number(
        serial_number, ctypes.byref(c_sensor_handle))
    return int(c_sensor_handle.value)


def get_sensor_information_by_index(sensor_index):
    c_sensor_info = cepton_sdk.c.C_SensorInformation()
    cepton_sdk.c.c_get_sensor_information_by_index(
        sensor_index, ctypes.byref(c_sensor_info))
    return SensorInformation.from_c(c_sensor_info)


def get_sensor_information_by_handle(sensor_handle):
    c_sensor_info = cepton_sdk.c.C_SensorInformation()
    cepton_sdk.c.c_get_sensor_information(
        sensor_handle, ctypes.byref(c_sensor_info))
    return SensorInformation.from_c(c_sensor_info)


def has_sensor(sensor_serial_number):
    try:
        get_sensor_handle(sensor_serial_number)
    except:
        return False
    return True


def get_sensor_information(sensor_serial_number):
    sensor_handle = get_sensor_handle(sensor_serial_number)
    return get_sensor_information_by_handle(sensor_handle)



import ctypes
import enum
import os.path
import uuid

import laspy
import plyfile
import numpy

import cepton_sdk.point


def save_points_las(points, path):
    """Save points to LAS file.
    """
    header = laspy.header.Header()
    header.data_format_id = 1

    with laspy.file.File(path, mode="w", header=header) as f:
        f.header.gps_time_type = 1
        f.header.guid = uuid.uuid1()
        f.header.scale = [0.001] * 3

        f.gps_time = numpy.remainder(points.timestamps, 1e9)
        f.x = points.positions[:, 0]
        f.y = points.positions[:, 1]
        f.z = points.positions[:, 2]
        f.intensity = points.intensities
        f.return_num = points.return_numbers


def load_points_las(load_path, cls=cepton_sdk.point.Points):
    """Load points from LAS file.

    Returns:
        Points
    """
    data = {}
    with laspy.file.File(load_path, mode="r") as f:
        points = cls(len(f.x))
        points.timestamps[:] = f.gps_time
        points.positions[:, 0] = f.x
        points.positions[:, 1] = f.y
        points.positions[:, 2] = f.z
        points.intensities[:] = f.intensity
        points.return_numbers[:] = f.return_num
    return points, data


class PlyPoint(ctypes.Structure):
    _fields_ = [
        ("position", 3 * ctypes.c_float),
    ]


def save_points_ply(points, path):
    """Save points to PLY file.
    """
    n = len(points)

    with open(path, "w") as f:
        lines = [
            "ply\n",
            "format binary_little_endian 1.0\n",
            "element vertex {}\n".format(len(points)),
            "property float x\n",
            "property float y\n",
            "property float z\n",
            "end_header\n",
        ]
        f.writelines(lines)

    dtype = numpy.dtype(PlyPoint)
    data = numpy.zeros([n], dtype=dtype)
    data["position"] = points.positions
    with open(path, "ab") as f:
        f.write(data.tobytes())


def load_points_ply(path):
    """Load points from PLY file.

    Returns:
        Points
    """
    plydata = plyfile.PlyData.read(input_path)
    data = plydata.elements[0].data
    n = len(data)
    points = cepton_sdk.Points(n)
    points.positions[:, 0] = data["x"]
    points.positions[:, 1] = data["y"]
    points.positions[:, 2] = data["z"]
    return points


class PcdPoint(ctypes.Structure):
    _fields_ = [
        ("position", 3 * ctypes.c_float),
    ]


def save_points_pcd(points, path):
    """Save points to PCD file.
    """
    n = len(points)

    with open(path, "w") as f:
        lines = [
            "VERSION 0.7\n",
            "FIELDS x y z\n",
            "SIZE 4 4 4\n",
            "TYPE F F F\n",
            "COUNT 1 1 1\n",
            "WIDTH {}\n".format(n),
            "HEIGHT 1\n",
            "VIEWPOINT 0 0 0 1 0 0 0\n",
            "POINTS {}\n".format(n),
            "DATA binary\n",
        ]
        f.writelines(lines)

    dtype = numpy.dtype(PcdPoint)
    data = numpy.zeros([n], dtype=dtype)
    data["position"] = points.positions
    with open(path, "ab") as f:
        f.write(data.tobytes())


def load_points_pcd(points):
    """Load points from PCD file.

    Returns:
        Points
    """
    raise NotImplementedError()


@enum.unique
class PointsFileType(enum.Enum):
    LAS = 1
    PCD = 2
    PLY = 3


def get_points_file_type(ext):
    return PointsFileType[ext.upper()]


def get_points_file_type_extension(file_type):
    return ".{}".format(file_type.name.lower())


def save_points(points, path, file_type=PointsFileType.LAS):
    if file_type == PointsFileType.LAS:
        save_points_las(points, path)
    elif file_type == PointsFileType.PCD:
        save_points_pcd(points, path)
    elif file_type == PointsFileType.PLY:
        save_points_ply(points, path)
    else:
        raise NotImplementedError()


def load_points(path, file_type=None):
    if file_type == None:
        ext = os.path.splitext(path)[1]
        file_type = get_points_file_type(ext)
    if file_type == PointsFileType.LAS:
        return load_points_las(path)
    elif file_type == PointsFileType.PCD:
        return load_points_pcd(path)
    elif file_type == PointsFileType.PLY:
        return load_points_ply(path)
    else:
        raise NotImplementedError()



import ctypes.util
import enum
import os
import os.path
import platform
import sys
import warnings
from ctypes import *

import numpy

SDK_VERSION = 11

# ------------------------------------------------------------------------------
# Load library
# ------------------------------------------------------------------------------


def _load_c_library():
    module_path = os.path.dirname(os.path.abspath(__file__))
    lib_dir = None
    lib_name = None
    if sys.platform.startswith("linux"):
        lib_dir = "lib/linux-{}/".format(platform.machine())
        lib_name = "libcepton_sdk.so".format()
    elif sys.platform.startswith("darwin"):
        lib_dir = "lib/osx/"
        lib_name = "libcepton_sdk.dylib"
    elif sys.platform.startswith("win"):
        lib_dir = "lib/win64/"
        lib_name = "cepton_sdk.dll"
    else:
        raise NotImplementedError("Platform not supported!")

    # Try local and global search paths
    module_file = os.path.join(module_path, lib_dir, lib_name)
    if not os.path.exists(module_file):
        module_file = lib_name
    return CDLL(module_file)


libcepton_sdk = _load_c_library()

# ------------------------------------------------------------------------------
# Arrays
# ------------------------------------------------------------------------------
_C_NDARRAY_REQUIREMENTS = ["C_CONTIGUOUS", "ALIGNED", "WRITEABLE", "OWNDATA"]


def get_c_ndpointer_type(dtype, ndim=1):
    return numpy.ctypeslib.ndpointer(
        dtype=dtype, ndim=ndim, flags=_C_NDARRAY_REQUIREMENTS)


def get_c_ndarray(a, dtype):
    return numpy.require(
        a, dtype=dtype, requirements=_C_NDARRAY_REQUIREMENTS)


def create_c_ndarray(size, dtype):
    a = numpy.zeros(size, dtype=dtype)
    return get_c_ndarray(a, dtype)


def convert_bytes_to_ndarray(a_bytes, c_type):
    """Convert bytes array to structured array"""
    dtype = numpy.dtype(c_type)
    assert (sizeof(c_type) == dtype.itemsize)
    a = a_bytes.view(dtype=dtype)
    assert (len(a) == len(a_bytes) / sizeof(c_type))
    return a


def convert_ndarray_to_bytes(a):
    """Convert structured array to bytes array"""
    return a.view(dtype=numpy.uint8)


def convert_c_array_to_ndarray(c_a, n, c_type):
    """Convert ctypes array to structured array"""
    n_bytes = n * sizeof(c_type)
    a_bytes = \
        numpy.ctypeslib.as_array(cast(c_a, POINTER(c_byte)), shape=(n_bytes,))
    return convert_bytes_to_ndarray(a_bytes, c_type)


def convert_ndarray_to_c_array(a, c_type):
    """Convert structured array to ctypes array"""
    a_bytes = convert_ndarray_to_bytes(a)
    assert (len(a_bytes) == len(a) * sizeof(c_type))
    c_a_bytes = numpy.ctypeslib.as_ctypes(a_bytes)
    c_a = cast(c_a_bytes, POINTER(c_type))
    return c_a


def unpackbits(a):
    bits = numpy.unpackbits(a.flatten().view(numpy.uint8)).astype(bool)
    bits = bits.reshape([a.size, -1, 8])
    bits = bits[:, :, ::-1]
    bits = bits.reshape(list(a.shape) + [-1])
    return bits


# ------------------------------------------------------------------------------
# General
# ------------------------------------------------------------------------------
C_SensorHandle = c_uint64


class C_ErrorCode(enum.IntEnum):
    CEPTON_SUCCESS = 0
    CEPTON_ERROR_GENERIC = -1
    CEPTON_ERROR_OUT_OF_MEMORY = -2
    CEPTON_ERROR_SENSOR_NOT_FOUND = -4
    CEPTON_ERROR_SDK_VERSION_MISMATCH = -5
    CEPTON_ERROR_COMMUNICATION = -6
    CEPTON_ERROR_TOO_MANY_CALLBACKS = -7
    CEPTON_ERROR_INVALID_ARGUMENTS = -8
    CEPTON_ERROR_ALREADY_INITIALIZED = -9
    CEPTON_ERROR_NOT_INITIALIZED = -10
    CEPTON_ERROR_INVALID_FILE_TYPE = -11
    CEPTON_ERROR_FILE_IO = -12
    CEPTON_ERROR_CORRUPT_FILE = -13
    CEPTON_ERROR_NOT_OPEN = -14
    CEPTON_ERROR_EOF = -15

    CEPTON_FAULT_INTERNAL = -1000
    CEPTON_FAULT_EXTREME_TEMPERATURE = -1001
    CEPTON_FAULT_EXTREME_HUMIDITY = -1002
    CEPTON_FAULT_EXTREME_ACCELERATION = -1003
    CEPTON_FAULT_ABNORMAL_FOV = -1004
    CEPTON_FAULT_ABNORMAL_FRAME_RATE = -1005
    CEPTON_FAULT_MOTOR_MALFUNCTION = -1006
    CEPTON_FAULT_LASER_MALFUNCTION = -1007
    CEPTON_FAULT_DETECTOR_MALFUNCTION = -1008


c_get_error_code_name = libcepton_sdk.cepton_get_error_code_name
c_get_error_code_name.argtypes = [c_int]
c_get_error_code_name.restype = c_char_p

c_is_error_code = libcepton_sdk.cepton_is_error_code
c_is_error_code.argtypes = [c_int]
c_is_error_code.restype = c_int

c_is_fault_code = libcepton_sdk.cepton_is_fault_code
c_is_fault_code.argtypes = [c_int]
c_is_fault_code.restype = c_int


def get_error_code_name(error_code):
    error_code_name = c_get_error_code_name(error_code)
    return error_code_name.decode("utf-8")


def is_error_code(error_code):
    return bool(c_is_error_code(error_code))


def is_fault_code(error_code):
    return bool(c_is_fault_code(error_code))


class C_Error(Exception):
    """Error thrown by most sdk functions.

    Attributes:
        error_code: `cepton_sdk.C_ErrorCode`
    """

    def __init__(self, message, error_code):
        super().__init__(message)
        self.error_code = error_code

    @classmethod
    def create(cls, error_code):
        message = get_error_code_name(error_code)
        return cls(message, error_code)


class C_Warning(Warning):
    pass


def check_error_code(error_code, warning=False):
    if not error_code:
        return
    if is_error_code(error_code) and (not warning):
        raise C_Error.create(error_code)
    else:
        message = get_error_code_name(error_code)
        warnings.warn(message, C_Warning, stacklevel=2)


def _c_error_check(error_code, func, args):
    check_error_code(error_code)


def _add_c_error_check(c_func):
    c_func.restype = c_int
    c_func.errcheck = _c_error_check


# ------------------------------------------------------------------------------
# Setup
# ------------------------------------------------------------------------------
class C_FrameOptions(Structure):
    _fields_ = [
        ("signature", c_size_t),
        ("mode", c_uint32),
        ("length", c_float),
    ]


c_create_frame_options = libcepton_sdk.cepton_sdk_create_frame_options
c_create_frame_options.restype = C_FrameOptions


class C_Options(Structure):
    _fields_ = [
        ("signature", c_size_t),
        ("control_flags", c_uint32),
        ("frame", C_FrameOptions),
        ("port", c_uint16),
    ]


c_create_options = libcepton_sdk.cepton_sdk_create_options
c_create_options.restype = C_Options

C_SensorErrorCallback = CFUNCTYPE(
    None, C_SensorHandle, c_int, c_char_p, c_void_p, c_size_t, c_void_p)

c_is_initialized = libcepton_sdk.cepton_sdk_is_initialized
c_is_initialized.restype = c_int

c_initialize = libcepton_sdk.cepton_sdk_initialize
c_initialize.argtypes = \
    [c_int, POINTER(C_Options), C_SensorErrorCallback, c_void_p]
_add_c_error_check(c_initialize)

c_deinitialize = libcepton_sdk.cepton_sdk_deinitialize
c_deinitialize.argtypes = []
_add_c_error_check(c_deinitialize)

c_set_control_flags = libcepton_sdk.cepton_sdk_set_control_flags
c_set_control_flags.argtypes = [c_uint32, c_uint32]
_add_c_error_check(c_set_control_flags)

c_get_control_flags = libcepton_sdk.cepton_sdk_get_control_flags
c_get_control_flags.argtypes = []
c_get_control_flags.restype = c_uint32

c_has_control_flag = libcepton_sdk.cepton_sdk_has_control_flag
c_has_control_flag.argtypes = [c_uint32]
c_has_control_flag.restype = c_int

c_get_port = libcepton_sdk.cepton_sdk_get_port
c_get_port.restype = c_uint16

c_set_port = libcepton_sdk.cepton_sdk_set_port
c_set_port.argtypes = [c_uint16]
_add_c_error_check(c_set_port)

c_set_frame_options = libcepton_sdk.cepton_sdk_set_frame_options
c_set_frame_options.argtypes = [POINTER(C_FrameOptions)]
_add_c_error_check(c_set_frame_options)

c_get_frame_mode = libcepton_sdk.cepton_sdk_get_frame_mode
c_get_frame_mode.restype = c_uint32

c_get_frame_length = libcepton_sdk.cepton_sdk_get_frame_length
c_get_frame_length.restype = c_float

c_clear_cache = libcepton_sdk.cepton_sdk_clear_cache
_add_c_error_check(c_clear_cache)

# ------------------------------------------------------------------------------
# Sensors
# ------------------------------------------------------------------------------


class C_SensorInformation(Structure):
    _fields_ = [
        ("handle", C_SensorHandle),
        ("serial_number", c_uint64),
        ("model_name", c_char * 28),
        ("model", c_int32),
        ("firmware_version", c_char * 32),

        ("last_reported_temperature", c_float),
        ("last_reported_humidity", c_float),
        ("last_reported_age", c_float),
        ("", c_float),

        ("ptp_ts", c_uint64),

        ("gps_ts_year", c_uint8),
        ("gps_ts_month", c_uint8),
        ("gps_ts_day", c_uint8),
        ("gps_ts_hour", c_uint8),
        ("gps_ts_min", c_uint8),
        ("gps_ts_sec", c_uint8),

        ("return_count", c_uint8),
        ("", c_uint8),

        ("is_mocked", c_uint32, 1),
        ("is_pps_connected", c_uint32, 1),
        ("is_nmea_connected", c_uint32, 1),
        ("is_ptp_connected", c_uint32, 1),
        ("is_calibrated", c_uint32, 1),
    ]


c_cepton_sensor_information_size = \
    c_size_t.in_dll(libcepton_sdk, "cepton_sensor_information_size").value
assert (sizeof(C_SensorInformation) == c_cepton_sensor_information_size)

c_get_n_sensors = libcepton_sdk.cepton_sdk_get_n_sensors
c_get_n_sensors.restype = c_size_t

c_get_sensor_handle_by_serial_number = \
    libcepton_sdk.cepton_sdk_get_sensor_handle_by_serial_number
c_get_sensor_handle_by_serial_number.argtypes = \
    [c_uint64, POINTER(C_SensorHandle)]
_add_c_error_check(c_get_sensor_handle_by_serial_number)

c_get_sensor_information = libcepton_sdk.cepton_sdk_get_sensor_information
c_get_sensor_information.argtypes = \
    [C_SensorHandle, POINTER(C_SensorInformation)]
_add_c_error_check(c_get_sensor_information)

c_get_sensor_information_by_index = \
    libcepton_sdk.cepton_sdk_get_sensor_information_by_index
c_get_sensor_information_by_index.argtypes = \
    [c_size_t, POINTER(C_SensorInformation)]
_add_c_error_check(c_get_sensor_information_by_index)

# ------------------------------------------------------------------------------
# Points
# ------------------------------------------------------------------------------


class C_SensorImagePoint(Structure):
    _fields_ = [
        ("timestamp", c_uint64),
        ("image_x", c_float),
        ("distance", c_float),
        ("image_z", c_float),
        ("intensity", c_float),
        ("return_number", c_uint8),
        ("valid", c_uint8),
        ("saturated", c_uint8),
        ("reserved", c_uint8)
    ]


c_cepton_sensor_image_point_size = \
    c_size_t.in_dll(libcepton_sdk, "cepton_sensor_image_point_size").value
assert (sizeof(C_SensorImagePoint) == c_cepton_sensor_image_point_size)

# ------------------------------------------------------------------------------
# Listen
# ------------------------------------------------------------------------------
C_SensorImageDataCallback = \
    CFUNCTYPE(None,
              C_SensorHandle, c_size_t, POINTER(C_SensorImagePoint), c_void_p)

c_listen_image_frames = libcepton_sdk.cepton_sdk_listen_image_frames
c_listen_image_frames.argtypes = [C_SensorImageDataCallback, c_void_p]
_add_c_error_check(c_listen_image_frames)

c_unlisten_image_frames = libcepton_sdk.cepton_sdk_unlisten_image_frames
_add_c_error_check(c_unlisten_image_frames)

# ------------------------------------------------------------------------------
# Capture
# ------------------------------------------------------------------------------
c_capture_replay_is_open = libcepton_sdk.cepton_sdk_capture_replay_is_open
c_capture_replay_is_open.restype = c_int

c_capture_replay_open = libcepton_sdk.cepton_sdk_capture_replay_open
c_capture_replay_open.argtypes = [c_char_p]
_add_c_error_check(c_capture_replay_open)

c_capture_replay_close = libcepton_sdk.cepton_sdk_capture_replay_close
_add_c_error_check(c_capture_replay_close)

c_capture_replay_get_start_time = \
    libcepton_sdk.cepton_sdk_capture_replay_get_start_time
c_capture_replay_get_start_time.restype = c_uint64

c_capture_replay_get_position = \
    libcepton_sdk.cepton_sdk_capture_replay_get_position
c_capture_replay_get_position.restype = c_float

c_capture_replay_get_length = \
    libcepton_sdk.cepton_sdk_capture_replay_get_length
c_capture_replay_get_length.restype = c_float

c_capture_replay_is_end = \
    libcepton_sdk.cepton_sdk_capture_replay_is_end
c_capture_replay_is_end.restype = c_int

c_capture_replay_rewind = \
    libcepton_sdk.cepton_sdk_capture_replay_rewind
_add_c_error_check(c_capture_replay_rewind)

c_capture_replay_seek = \
    libcepton_sdk.cepton_sdk_capture_replay_seek
c_capture_replay_seek.argtypes = [c_float]
_add_c_error_check(c_capture_replay_seek)

c_capture_replay_get_enable_loop = \
    libcepton_sdk.cepton_sdk_capture_replay_get_enable_loop
c_capture_replay_get_enable_loop.restype = c_int

c_capture_replay_set_enable_loop = \
    libcepton_sdk.cepton_sdk_capture_replay_set_enable_loop
c_capture_replay_set_enable_loop.argtypes = [c_bool]
_add_c_error_check(c_capture_replay_set_enable_loop)

c_capture_replay_get_speed = \
    libcepton_sdk.cepton_sdk_capture_replay_get_speed
c_capture_replay_get_speed.restype = c_float

c_capture_replay_set_speed = \
    libcepton_sdk.cepton_sdk_capture_replay_set_speed
c_capture_replay_set_speed.argtypes = [c_float]
_add_c_error_check(c_capture_replay_set_speed)

c_capture_replay_resume_blocking_once = \
    libcepton_sdk.cepton_sdk_capture_replay_resume_blocking_once
_add_c_error_check(c_capture_replay_resume_blocking_once)

c_capture_replay_resume_blocking = \
    libcepton_sdk.cepton_sdk_capture_replay_resume_blocking
c_capture_replay_resume_blocking.argtypes = [c_float]
_add_c_error_check(c_capture_replay_resume_blocking)

c_capture_replay_is_running = \
    libcepton_sdk.cepton_sdk_capture_replay_is_running
c_capture_replay_is_running.restype = c_int

c_capture_replay_resume = libcepton_sdk.cepton_sdk_capture_replay_resume
_add_c_error_check(c_capture_replay_resume)

c_capture_replay_pause = libcepton_sdk.cepton_sdk_capture_replay_pause
_add_c_error_check(c_capture_replay_pause)



import platform
import sys

import setuptools

if __name__ == "__main__":
    setuptools.setup(
        name="cepton_sdk",
        version="0.2.1",
        description="Cepton Python SDK",
        long_description=open("README.md").read(),
        url="https://github.com/ceptontech/cepton_sdk_redist",
        author="Cepton Technologies",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Programming Language :: Python :: 3",
        ],
        keywords="cepton sdk",
        python_requires=">=3.3",
        packages=setuptools.find_packages(),
        include_package_data=True,
        install_requires=[
            "numpy",
            "pyqt5",
            "pyserial",
            "transforms3d",
            "uuid",
        ]
    )



import cepton_sdk.c
import numpy
from cepton_sdk.common import *

__all__ = [
    "ImagePoints",
    "Points",
]


class Points(StructureOfArrays):
    """3D points array.

    Attributes:
        timestamps: time since epoch [microseconds]
        positions: x, y, z cartesian coordinates
        intensities: 0-1 scaled intensity
        return_numbers
        valid
    """

    def __init__(self, n=0):
        super().__init__(n)
        self.timestamps = numpy.zeros([n])
        self.positions = numpy.zeros([n, 3])
        self.intensities = numpy.zeros([n])
        self.return_numbers = numpy.zeros([n], dtype=numpy.uint8)
        self.valid = numpy.zeros([n], dtype=bool)
        self.saturated = numpy.zeros([n], dtype=bool)
        self._check_array_members()

    @classmethod
    def _get_array_member_names(cls):
        return ["timestamps", "positions", "intensities", "return_numbers",
                "valid", "saturated"]


class ImagePoints(StructureOfArrays, ToCMixin):
    """Image points array.

    Attributes:
        timestamps: time since epoch [microseconds]
        positions: x, z image coordinates
        distances: distance [meters]
        intensities: 0-1 scaled intensity
        return_numbers
        valid
    """

    def __init__(self, n=0):
        super().__init__(n)
        self.timestamps = numpy.zeros([n])
        self.positions = numpy.zeros([n, 2])
        self.distances = numpy.zeros([n])
        self.intensities = numpy.zeros([n])
        self.return_numbers = numpy.zeros([n], dtype=numpy.uint8)
        self.valid = numpy.zeros([n], dtype=bool)
        self.saturated = numpy.zeros([n], dtype=bool)
        self._check_array_members()

    @classmethod
    def _get_c_class(cls):
        return cepton_sdk.c.C_SensorImagePoint

    @classmethod
    def _get_array_member_names(cls):
        return ["timestamps", "positions", "distances", "intensities",
                "return_numbers", "valid", "saturated"]

    @classmethod
    def from_c(cls, n_points, c_image_points):
        data = \
            cepton_sdk.c.convert_c_array_to_ndarray(
                c_image_points, n_points, cls._get_c_class())

        image_points = cls(n_points)
        image_points.timestamps[:] = 1e-6 * data["timestamp"].astype(float)
        image_points.positions[:, 0] = data["image_x"]
        image_points.positions[:, 1] = data["image_z"]
        image_points.distances[:] = data["distance"]
        image_points.intensities[:] = data["intensity"]
        image_points.return_numbers[:] = data["return_number"]
        image_points.valid[:] = data["valid"]
        image_points.saturated[:] = data["saturated"]
        return image_points

    def to_c(self, c_type=None):
        if c_type is None:
            c_type = self._get_c_class()

        dtype = numpy.dtype(c_type)
        data = numpy.zeros(len(self), dtype=dtype)
        data["timestamp"][:] = (1e6 * self.timestamps).astype(numpy.uint64)
        data["image_x"][:] = self.positions[:, 0]
        data["image_z"][:] = self.positions[:, 1]
        data["distance"][:] = self.distances
        data["intensity"][:] = self.intensities
        data["return_number"][:] = self.return_numbers
        data["valid"][:] = self.valid
        data["saturated"][:] = self.saturated

        c_image_points_ptr = \
            cepton_sdk.c.convert_ndarray_to_c_array(data, c_type)
        return c_image_points_ptr

    def to_points(self, cls=Points):
        """Convert to 3d points.
        """
        assert issubclass(cls, Points)

        points = cls(len(self))
        points.timestamps[:] = self.timestamps
        points.intensities[:] = self.intensities
        points.return_numbers[:] = self.return_numbers
        points.valid[:] = self.valid
        points.saturated[:] = self.saturated

        hypotenuse_small = numpy.sqrt(
            self.positions[:, 0]**2 +
            self.positions[:, 1]**2 + 1)
        ratio = self.distances / hypotenuse_small
        points.positions[:, 0] = -self.positions[:, 0] * ratio
        points.positions[:, 1] = ratio
        points.positions[:, 2] = -self.positions[:, 1] * ratio
        return points

    @classmethod
    def from_points(cls, points):
        """Create from 3d points.
        """
        assert isinstance(points, Points)

        image_points = cls(len(points))
        image_points.timestamps[:] = points.timestamps
        image_points.intensities[:] = points.intensities
        image_points.return_numbers[:] = points.return_numbers
        image_points.valid[:] = points.valid
        image_points.saturated[:] = points.saturated

        image_points.positions[:, :] = \
            -points.positions[:, [0, 2]] / points.positions[:, [1]]
        image_points.distances = numpy.sqrt(
            points.positions[:, 0]**2 +
            points.positions[:, 1]**2 +
            points.positions[:, 2]**2)
        return image_points


class NPointsAccumulator(object):
    """Points accumulator used by internal listeners."""

    def __init__(self, ):
        self._points = None
        self.reset()

    def reset(self):
        if self._points is not None:
            cls = type(self._points)
            self._points = cls(0)

    def add_points(self, points):
        if len(points) == 0:
            return

        if self._points is None:
            self._points = points
        else:
            self._points = self._points.combine([self._points, points])

    def has_points(self, n):
        if self._points is None:
            return False
        return len(self._points) >= n

    def get_points(self, n):
        if self._points is None:
            return None
        points = self._points[:n]
        self._points = self._points[n:]
        return points


class _TimePointsAccumulatorBase(object):
    """Points accumulator used by internal listeners."""

    def __init__(self, t=-numpy.inf):
        self.reset()
        self._t = t

    def reset(self):
        self._t = -numpy.inf

    @property
    def t(self):
        return self._t

    def _check_t(self, t):
        if t <= self._t:
            raise ValueError("requested times must be increasing")

    def is_empty(self):
        return numpy.isinf(self._t)


class TimePointsAccumulator(_TimePointsAccumulatorBase):
    """Points accumulator used by internal listeners."""

    def __init__(self, **kwargs):
        self._points = None
        super().__init__(**kwargs)

    @property
    def t_max(self):
        return self._t_max

    def reset(self):
        super().reset()
        self._t_max = -numpy.inf
        if self._points is not None:
            cls = type(self._points)
            self._points = cls(0)

    def add_points(self, points):
        if len(points) == 0:
            return

        if numpy.isinf(self._t):
            self._t = numpy.amin(points.timestamps)
        else:
            select = points.timestamps > self._t
            points = points[select]
            if len(points) == 0:
                return

        if self._points is None:
            self._points = points
        else:
            self._points = self._points.combine([self._points, points])

        # Sort
        sort_indices = numpy.argsort(self._points.timestamps)
        self._points = self._points[sort_indices]

        self._t_max = self._points.timestamps[-1]

    def has_points_by_t(self, t):
        if self.is_empty():
            return False
        self._check_t(t)
        return t <= self._t_max

    def has_points(self, t_length):
        t = self._t + t_length
        return self.has_points_by_t(t)

    def get_points_by_t(self, t):
        self._check_t(t)
        if self._points is None:
            points = None
        else:
            is_valid = self._points.timestamps <= t
            points = self._points[is_valid]

        self._t = t
        if self._points is not None:
            is_valid = self._points.timestamps > self._t
            self._points = self._points[is_valid]

        return points

    def get_points(self, t_length):
        t = self._t + t_length
        return self.get_points_by_t(t)


class MultipleTimePointsAccumulator(_TimePointsAccumulatorBase):
    """Points accumulator used by internal listeners."""

    def __init__(self, t_latency=0, **kwargs):
        self._t_latency = t_latency
        super().__init__(**kwargs)

    @property
    def t_max(self):
        return max(
            map(lambda x: x.t_max, self._accumulators.values()),
            default=-numpy.inf)

    def reset(self):
        super().reset()
        self._accumulators = {}

    def add_points(self, serial_number, points):
        if len(points) == 0:
            return

        if numpy.isinf(self._t):
            self._t = numpy.amin(points.timestamps) + self._t_latency

        if serial_number not in self._accumulators:
            options = {
                "t": self._t,
            }
            self._accumulators[serial_number] = \
                TimePointsAccumulator(**options)
        self._accumulators[serial_number].add_points(points)

    def has_points_by_t(self, t):
        if self.is_empty():
            return False
        self._check_t(t)
        for accumulator in self._accumulators.values():
            if accumulator.has_points_by_t(t + self._t_latency):
                return True
        return False

    def has_points(self, t_length):
        t = self._t + t_length
        return self.has_points_by_t(t)

    def get_points_by_t(self, t):
        self._check_t(t)
        self._t = t
        points_dict = {}
        for serial_number, accumulator in self._accumulators.items():
            points = accumulator.get_points_by_t(t)
            if (points is None) or (len(points) == 0):
                continue
            points_dict[serial_number] = points
        return points_dict

    def get_points(self, t_length):
        t = self._t + t_length
        return self.get_points_by_t(t)
